/***************************************************************************
 *   Copyright (C) 2009 by Miguel Angel Rodriguez Jodar                    *
 *   rodriguj@atc.us.es                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/*
CHANGELOG
---------

2009-Apr-13: (rodriguj@atc.us.es) 
WinAmp TAP input plugin ported to a standalone Linux 
console application. First version.

TODO: probably, a lot.
BUGS: surely, a lot.
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <fcntl.h>
#ifdef SDL
#include <SDL/SDL.h>
#include <SDL/SDL_mixer.h>
#else
#include <linux/soundcard.h>
#include <sys/ioctl.h>
#endif


#define PI 3.141592654

enum {SQUARE=0, SINE};
enum {INICIO=0, SILENCIO, TONOGUIA, SINCRO, DATOS, TONOFINAL};

int sfreq=44100;
int waverender=SQUARE;
int endtone=0;
int pausebtracks=1;
int quietmode=0;

int f_leader=800;
int f_bit1=1000;
int f_bit0=2000;
int f_synchr=2500;
int f_endtone=500;

char filename[256];
int input_file;

int Estado=INICIO;
int teSilencio;
int teTonoGuia, ttTonoGuia;
int teSincro, ttSincro;
int teBit, ttBit;
int teTonoFinal, ttTonoFinal;
int DatoActual, PosicionBit, ValorBit, ContBits;
int muestraactual, byteactual;
int lbloque;

short int *SenalTonoGuia; // un periodo del tono guia (leader)
short int *SenalBit1;  // un periodo de un bit 1
short int *SenalBit0;  // un periodo de un bit 0
short int *SenalSincro; // un periodo del pulso de sincronismo
short int *SenalTonoFinal;  // un periodo del tono marca del final
char *Bloque;  // the TAP block currently being processed.

void ShowHeader (char *hdr)
{
  int i;
  char type;
  unsigned short lblk, par1, par2;
  
  type=hdr[1];
  lblk = *(short int *)(hdr+12);
  par1 = *(short int *)(hdr+14);
  par2 = *(short int *)(hdr+16);
  
  switch (type)  // type of header
  {
    case 0:
      printf ("Program: ");
      break;
    case 1:
      printf ("Number array: ");
      break;
    case 2:
      printf ("Character array: ");
      break;
    case 3:
      printf ("Bytes:   ");
      break;
    default:
      printf ("Unknown type %02.2Xh: ", type);
      break;
  }
  
  for (i=2;i<=11;i++)
    putchar(hdr[i]);
  
  printf ("  Length: %d  ", lblk);
  
  if (type==0)
  {
    printf ("Autostart: ");
    if (par1<32768)
      printf ("%d  ", par1);
    else
      printf ("none  ");
    printf ("Vars at: %d", par2);
  }
  else if (type==3)
    printf ("Start: %d", par1);
  
  printf ("                    \n");
}

void GenerarPeriodoCuadrado (short int *buffer, int l)
{
  int i;

  for (i=0;i<l;i++)
    if (i<(l/2))
      buffer[i]=-32760;
  else
    buffer[i]=32760;
}

void GenerarPeriodoSeno (short int *buffer, int l)
{
  int i;

  for (i=0;i<l;i++)
    buffer[i]=(short int)(-32760*sin(2*PI*i/l));
}

void GenerarSenales (void)
{
  if (waverender==SQUARE)
  {
    GenerarPeriodoCuadrado (SenalTonoGuia, sfreq/f_leader);
    GenerarPeriodoCuadrado (SenalBit1, sfreq/f_bit1);
    GenerarPeriodoCuadrado (SenalBit0, sfreq/f_bit0);
    GenerarPeriodoCuadrado (SenalSincro, sfreq/f_synchr);
    GenerarPeriodoCuadrado (SenalTonoFinal, sfreq/f_endtone);
  }
  else
  {
    GenerarPeriodoSeno (SenalTonoGuia, sfreq/f_leader);
    GenerarPeriodoSeno (SenalBit1, sfreq/f_bit1);
    GenerarPeriodoSeno (SenalBit0, sfreq/f_bit0);
    GenerarPeriodoSeno (SenalSincro, sfreq/f_synchr);
    GenerarPeriodoSeno (SenalTonoFinal, sfreq/f_endtone);
  }
}

void IniciarPlugin (void)
{
  SenalTonoGuia = malloc(sfreq*sizeof(short int)/f_leader);
  SenalBit1 = malloc(sfreq*sizeof(short int)/f_bit1);
  SenalBit0 = malloc(sfreq*sizeof(short int)/f_bit0);
  SenalSincro = malloc(sfreq*sizeof(short int)/f_synchr);
  SenalTonoFinal = malloc(sfreq*sizeof(short int)/f_endtone);
  Bloque = malloc(65536);
  
  GenerarSenales();
}

int LeerBloque (char *bloque)
{
  short int lb;
  int leido;

  leido = read (input_file, &lb, 2);
  if (leido!=2)
    return 0;

  leido = read (input_file, bloque, lb);
  if (leido!=lb)
    return 0;

  return lb;
}

int Render576 (short int *buffer)
{
  int i;
  unsigned char flag;

  for (i=0;i<576;i++)
    buffer[i]=0;

  for (i=0;i<576;i++)
  {
    switch (Estado)
    {
      case INICIO:
        lbloque=LeerBloque(Bloque);
        if (lbloque==0)                 // si no hay mas bloque, se acabo
        {
          if (endtone)
          {
            teTonoFinal=0;
            ttTonoFinal=3*sfreq;
            Estado=TONOFINAL;
            muestraactual=0;
            break;
          }
          else
            return 0;
        }
        Estado=TONOGUIA;
        teTonoGuia=0;
        muestraactual=0;
        byteactual=0;
        flag=Bloque[byteactual];
        if (flag&0x80)                 // la duración del tono guia la da el flag de grabacion
          ttTonoGuia=2*sfreq;    // duración tono guia corto: 2 s
        else
          ttTonoGuia=5*sfreq;    // duración tono guía largo: 5 s
        break;

      case SILENCIO:
        if (Bloque[0]==0x00 && lbloque==19 && quietmode==0 && teSilencio==0)
          ShowHeader(Bloque);
        buffer[i]=0;
        teSilencio++;   // tiempo en ms que dura una muestra
        if (teSilencio>sfreq || (pausebtracks==0 && teSilencio>(sfreq/10)))  // tiempo de silencio: 1000 ms (conf.) o 0.1s si no hay pausas
          Estado=INICIO;
        break;

      case TONOGUIA:
        buffer[i]=SenalTonoGuia[muestraactual];
        muestraactual=(muestraactual>=(sfreq/f_leader))? 0 : muestraactual+1;
        teTonoGuia++;
        if (teTonoGuia>=ttTonoGuia && muestraactual==0)
        {
          Estado=SINCRO;
          teSincro=0;
          muestraactual=0;
          ttSincro=sfreq/f_synchr;
        }
        break;

      case SINCRO:
        buffer[i]=SenalSincro[muestraactual];
        muestraactual=(muestraactual>=(sfreq/f_synchr))? 0 : muestraactual+1;
        teSincro++;
        if (teSincro>=ttSincro)
        {
          Estado=DATOS;
          DatoActual=Bloque[byteactual];
          PosicionBit=128;
          ContBits=0;
          ValorBit= (DatoActual&PosicionBit)? 1 : 0;
          muestraactual=0;
          teBit=0;
          ttBit=sfreq/((ValorBit==0)? f_bit0 : f_bit1);   //0 -> 2000 Hz, 1 -> 1000 Hz
        }
        break;

      case DATOS:
        buffer[i]=(ValorBit==0)? SenalBit0[muestraactual] : SenalBit1[muestraactual];
        muestraactual++;
        teBit++;
        if (teBit>=ttBit)
        {
          PosicionBit/=2;
          ContBits++;
          if (ContBits==8)
          {
            PosicionBit=128;
            ContBits=0;
            byteactual++;
            if (byteactual>=lbloque)
            {
              Estado=SILENCIO;
              teSilencio=0;
              break;
            }
            DatoActual=Bloque[byteactual];
          }
          ValorBit= (DatoActual&PosicionBit)? 1 : 0;
          muestraactual=0;
          teBit=0;
          ttBit=sfreq/((ValorBit==0)? f_bit0 : f_bit1);
        }
        break;

      case TONOFINAL:
        if (teTonoFinal<sfreq)
          buffer[i]=0;
        else
          buffer[i]=SenalTonoFinal[muestraactual];
        muestraactual=(muestraactual>=(sfreq/f_endtone))? 0 : muestraactual+1;
        teTonoFinal++;
        if (teTonoFinal>=ttTonoFinal)
          return 0;
        break;
    }
  }
  return 576*2;
}

#ifndef SDL
void send_ioctl_dsp (int fd, int cmd, int arg, char *ncmd, char *merror)
{
  int tmp=arg;
      
  if (ioctl (fd, cmd, &tmp) == -1)
  {
    perror (ncmd);
    exit (EXIT_FAILURE);
  }
  if (tmp != arg)
  {
    fprintf (stderr, merror);
    exit (EXIT_FAILURE);
  }
}
#endif

int main(int argc, char *argv[])
{
  int i;
  int fd;
  int canal=-1;
  short int *samples;
  #ifdef SDL
  Mix_Chunk chunk;
  #endif
  
  if (argc<2)
  {
    fprintf (stderr, "Use -h to show usage and available options.\n");
    return EXIT_FAILURE;
  }
  
  for (i=1;i<argc;i++)
  {
    if (argv[i][0]=='-')
    {
      switch (tolower(argv[i][1]))
      {
        case 'f':
          sfreq=atoi(&argv[i][2]);
          break;
          
        case 'r':
          waverender=(tolower(argv[i][2])=='s')? SQUARE : SINE;
          break;
          
        case 'l':
          f_leader=atoi(&argv[i][2]);
          break;
        
        case 's':
          f_synchr=atoi(&argv[i][2]);
          break;
        
        case 'o':
          f_bit1=atoi(&argv[i][2]);
          break;
        
        case 'z':
          f_bit0=atoi(&argv[i][2]);
          break;
        
        case 'm':
          f_endtone=atoi(&argv[i][2]);
          break;
        
        case 'p':
          pausebtracks=(argv[i][2]=='0')? 0 : 1;
          break;
          
        case 'e':
          endtone=(argv[i][2]=='0')? 0 : 1;
          break;
          
        case 'q':
          quietmode=1;
          break;
        
        case 'h':
        case '?':
          fprintf (stderr, "Usage: playtap [options] file[.tap]\n");
          fprintf (stderr, " OPTIONS:\n");
          fprintf (stderr, "   -fNNNN  : sets sampling frequency of resulting audio signal (def. to 44100 Hz)\n");
          fprintf (stderr, "   -rs     : renders pulses using square waves (default).\n");
          fprintf (stderr, "   -rw     : renders pulses using sine waves.\n");
          fprintf (stderr, "   -lNNNN  : sets frequency for pilot tone (def. %d Hz).\n", f_leader);
          fprintf (stderr, "   -sNNNN  : sets frequency for synchr pulse (def. %d Hz).\n", f_synchr);
          fprintf (stderr, "   -oNNNN  : sets frequency for bit 1 (def. %d Hz).\n", f_bit1);
          fprintf (stderr, "   -zNNNN  : sets frequency for bit 0 (def. %d Hz).\n", f_bit0);
          fprintf (stderr, "   -mNNNN  : sets frequency for ending tone (only if -e1 used, def. %d Hz\n", f_endtone);
          fprintf (stderr, "   -p0|1   : if 1, inserts a 1 sec pause betwween blocks (def. enabled).\n");
          fprintf (stderr, "   -e0|1   : if 1, issues a tone to mark the end of playing (def. disabled).\n");
          fprintf (stderr, "   -q      : Disables status and information (def. enabled).\n");
          fprintf (stderr, "   -h -?   : shows this help screen.\n\n");
          return EXIT_SUCCESS;
          break;
          
        default:
          fprintf (stderr, "Option not recognized. Use -h to get all the available options.\n");
          return EXIT_FAILURE;
      }
    }
    else
      strncpy (filename, argv[i], 250);
  }
      
  if (strncasecmp(filename+strlen(filename)-4, ".tap", 4)!=0)
    strcat (filename, ".tap");
  
#ifdef SDL
  if(SDL_Init(SDL_INIT_AUDIO) != 0)
  {
     fprintf(stderr, "Unable to initialize audio: %s\n", SDL_GetError());
     exit(EXIT_FAILURE);
  }
  if(Mix_OpenAudio(sfreq, AUDIO_S16, 1, 576))
  {
     fprintf(stderr, "Audio could not be set up for %d Hz 16 bit mono: ",
       sfreq, SDL_GetError());
     exit(EXIT_FAILURE);
  }
#else
  if ((fd = open ("/dev/dsp", O_WRONLY, 0)) == -1)
  {
    perror ("open /dev/dsp");
    exit (EXIT_FAILURE);
  }

  send_ioctl_dsp (fd, SNDCTL_DSP_SETFMT, AFMT_S16_LE, "SNDCTL_DSP_SETFMT", "This soundcard doesn't accept 16 bit samples, little endian format. Exiting...\n");
  send_ioctl_dsp (fd, SNDCTL_DSP_CHANNELS, 1, "SNDCTL_DSP_NCHANNELS", "This soundcard doesn't accept mono samples. Exiting...\n");
  send_ioctl_dsp (fd, SNDCTL_DSP_SPEED, sfreq, "SNDCTL_DSP_SPEED", "This soundcard doesn't accept the programmed sampling frequency. Choose another. Exiting...\n");
#endif
      
  input_file = open (filename, O_RDONLY, 0);
  if (input_file<0)
  {
    perror("Trying to open TAP file");
    exit (EXIT_FAILURE);
  }
 
  samples = malloc(576*sizeof(short int));
  IniciarPlugin();
  
  if (!quietmode)
  {
    printf ("\nPlayTAP v0.1\n");
    printf ("(C)1998-2009 Miguel Angel Rodriguez Jodar (http://www.zxprojects.com)\n");
    printf ("GPL Licensed.\n");
    printf ("-------------- PARAMETERS --------------\n");
    printf (" Sampling frequency:   %d Hz\n", sfreq);
    printf (" Leader tone:          %d Hz\n", f_leader);
    printf (" Sinc pulse:           %d Hz\n", f_synchr);
    printf (" Bit 0:                %d Hz\n", f_bit0);
    printf (" Bit 1:                %d Hz\n", f_bit1);
    printf (" EOP mark tone:        %d Hz\n", f_endtone);
    printf (" Rendering:            %s\n", (waverender==SQUARE)? "square wave" : "sine wave");
    printf (" Pause between blocks: %s\n", (pausebtracks)? "YES" : "NO");
    printf (" Issue EOP mark tone:  %s\n", (endtone)? "YES" : "NO");
    printf ("-----------------------------------------\n");
       
    printf ("%s is loading. Please, wait.\n\n", filename);
  }        
 
#ifdef SDL
  chunk.volume=MIX_MAX_VOLUME;
  chunk.abuf=(Uint8 *)samples;
  chunk.alen=576*2;
  chunk.allocated=1;
#endif

  while (Render576(samples)==(576*2))
  {
#ifdef SDL
    if(canal > -1)
       while(Mix_Playing(canal) != 0);
    canal=Mix_PlayChannel(-1, &chunk, 0);
    if(canal == -1)
    {
       fprintf(stderr, "Unable to play sound: %s\n", Mix_GetError());
       exit(EXIT_FAILURE);
    }
#else
    write (fd, samples, 576*2);
#endif
    if (!quietmode && Estado==DATOS && lbloque>50)
    {
      putchar ('[');
      for (i=0;i<=(byteactual*50/lbloque);i++)
        putchar('=');
      for (;i<50;i++)
        putchar(' ');
      printf ("]       \r");
      fflush (stdout);
    }
  }

  printf ("\n");
  
#ifdef SDL
  Mix_CloseAudio();
  SDL_Quit();
#else
  close (fd);
#endif
  close (input_file);
  
  return EXIT_SUCCESS;
}
